// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

var DotNetEntropyLib = {
    $DOTNETENTROPY: {
        // batchedQuotaMax is the max number of bytes as specified by the api spec.
        // If the byteLength of array is greater than 65536, throw a QuotaExceededError and terminate the algorithm.
        // https://www.w3.org/TR/WebCryptoAPI/#Crypto-method-getRandomValues
        batchedQuotaMax: 65536,
        getBatchedRandomValues: function (buffer, bufferLength) {
            // for modern web browsers
            // map the work array to the memory buffer passed with the length
            for (var i = 0; i < bufferLength; i += this.batchedQuotaMax) {
                var view = new Uint8Array(Module.HEAPU8.buffer, buffer + i, Math.min(bufferLength - i, this.batchedQuotaMax));
                crypto.getRandomValues(view)
            }
        }
    },
    dotnet_browser_entropy : function (buffer, bufferLength) {
        // check that we have crypto available
        if (typeof crypto === 'object' && typeof crypto['getRandomValues'] === 'function') {
            DOTNETENTROPY.getBatchedRandomValues(buffer, bufferLength)
            return 0;
        } else {
            // we couldn't find a proper implementation, as Math.random() is not suitable
            // instead of aborting here we will return and let managed code handle the message
            return -1;
        }
    },
};

autoAddDeps(DotNetEntropyLib, '$DOTNETENTROPY')
mergeInto(LibraryManager.library, DotNetEntropyLib)

// SIG // Begin signature block
// SIG // MIIokgYJKoZIhvcNAQcCoIIogzCCKH8CAQExDzANBglg
// SIG // hkgBZQMEAgEFADB3BgorBgEEAYI3AgEEoGkwZzAyBgor
// SIG // BgEEAYI3AgEeMCQCAQEEEBDgyQbOONQRoqMAEEvTUJAC
// SIG // AQACAQACAQACAQACAQAwMTANBglghkgBZQMEAgEFAAQg
// SIG // He0Yy7K+G59QXMN5sHHtOtGJyoFg0Ie3YpqLi1MNoKeg
// SIG // gg30MIIGcjCCBFqgAwIBAgITMwAAA3uhCj7LZukBwAAA
// SIG // AAADezANBgkqhkiG9w0BAQwFADB+MQswCQYDVQQGEwJV
// SIG // UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMH
// SIG // UmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBv
// SIG // cmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBT
// SIG // aWduaW5nIFBDQSAyMDExMB4XDTIzMDUxMTE5MDMzMFoX
// SIG // DTI0MDUwODE5MDMzMFowYzELMAkGA1UEBhMCVVMxEzAR
// SIG // BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
// SIG // bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
// SIG // bjENMAsGA1UEAxMELk5FVDCCAaIwDQYJKoZIhvcNAQEB
// SIG // BQADggGPADCCAYoCggGBAJOpuFVvwZX8es6IIHp1wzKi
// SIG // TB6W/bNlhAg2RxyOXmayw41E23NuahOPrFMSjBFVLPL5
// SIG // QfUjA2L/Iif0dtGA5TYvZRsR3mln6tgqZnCD8ECCzUKI
// SIG // dQYFJzO5Y2koNLSLDvSPAiR8xsU/nXfIpThFOiMDfXnr
// SIG // GmbiFfERnf691t2gy2jB/Oawu6sS5Kyqzbghx8pnuHab
// SIG // NAjHxxsxxzjcFL/2uOByCtFQbGeVKnEDTjyuon4sYabJ
// SIG // u8BC6lkmAim6tjN7hc58ZHK9S3+3CeRxVGxxtszbWk8u
// SIG // j/4TNh9wS8Coy824lCLv7GplBDLFN1D8/ADCSwBfQGfg
// SIG // sh02BwIB1Bra+7plEXFzAeycYL4QiMlyEEWmxBMO5YIh
// SIG // JiLIFJBC7s580kO8IYtk+d5yAUPLtSIC2SHj7dyzz4dP
// SIG // CY9VS2H69CvXi4W0KX6pLlkhdHLEQo/Achl9jVxp29i2
// SIG // gR3VMZgkTIS2qYFpdof6f4jrtMXbOH9Xu9+cEAnBEhxd
// SIG // OqIscwIDAQABo4IBgjCCAX4wHwYDVR0lBBgwFgYKKwYB
// SIG // BAGCN0wIAQYIKwYBBQUHAwMwHQYDVR0OBBYEFL7FcgNC
// SIG // ZG2sGuJAz92BD/oS98C4MFQGA1UdEQRNMEukSTBHMS0w
// SIG // KwYDVQQLEyRNaWNyb3NvZnQgSXJlbGFuZCBPcGVyYXRp
// SIG // b25zIExpbWl0ZWQxFjAUBgNVBAUTDTQ2NDIyMys1MDEw
// SIG // MzMwHwYDVR0jBBgwFoAUSG5k5VAF04KqFzc3IrVtqMp1
// SIG // ApUwVAYDVR0fBE0wSzBJoEegRYZDaHR0cDovL3d3dy5t
// SIG // aWNyb3NvZnQuY29tL3BraW9wcy9jcmwvTWljQ29kU2ln
// SIG // UENBMjAxMV8yMDExLTA3LTA4LmNybDBhBggrBgEFBQcB
// SIG // AQRVMFMwUQYIKwYBBQUHMAKGRWh0dHA6Ly93d3cubWlj
// SIG // cm9zb2Z0LmNvbS9wa2lvcHMvY2VydHMvTWljQ29kU2ln
// SIG // UENBMjAxMV8yMDExLTA3LTA4LmNydDAMBgNVHRMBAf8E
// SIG // AjAAMA0GCSqGSIb3DQEBDAUAA4ICAQB+Zbp5xRWoPCFA
// SIG // xpfmeVgVuYxeHBATOkt36f7cQvvyX/dlCOTptS7gJqvW
// SIG // qPyWuZSNq9gwSWrHHF/BnjohTia2lLphcFRNmeXo8THS
// SIG // 8+Fl6FzyjRTc+BWNgdee/MzKYVv8bnAMoRE/Dl5Fdjzl
// SIG // JCtq0I/2zlKLjuBbTb9BsMdlBaLKNzq9wZQGOkXxfcet
// SIG // 0txDHxFfhtKJr6GPIAl7BokhJeTBtsrGicMXlYDACNHf
// SIG // Nz7jHbgDHbuSbf2p29WruXCrIvg29zDrj2CnGJfTWLhY
// SIG // 48GZh1h88OeCkufx4YUba2FSUV+QPoPtYOCrmCFL9lmg
// SIG // vIhds1oAibxMDI/YwLV9//Jd9VILgTvWZfv22PyuXnIM
// SIG // /kNEHVExKj7UBqx3JmHwh/AOeUia1ihWkJEP2/J9JNhn
// SIG // n2A2eu7FFzYhAQYLhQ01FajGdixk7TqVlpA8Ofm1t+Pw
// SIG // cilbnM9ZEbP5y/HN1Lu35SOhDCPx9nyhBcQda0CB0O8k
// SIG // gFq0U/q7FBIGwNG7/gJ535bIaAGS/A6RUMBeHqBqMBQk
// SIG // khrFPx3Vsfq75WrHzCnMsKcA7tG5ZyV5ld6SbJ2ddNUv
// SIG // iajw3lcR5GL6ECdQB8jDqMNE1eJOmQ9kynznLEGbCHRl
// SIG // M7Pu5rf7QXn1h1GWmIYMy5WMcH1323Fh1RaegqARdyQL
// SIG // hdiE5kXk6jCCB3owggVioAMCAQICCmEOkNIAAAAAAAMw
// SIG // DQYJKoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMw
// SIG // EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRt
// SIG // b25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRp
// SIG // b24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENlcnRp
// SIG // ZmljYXRlIEF1dGhvcml0eSAyMDExMB4XDTExMDcwODIw
// SIG // NTkwOVoXDTI2MDcwODIxMDkwOVowfjELMAkGA1UEBhMC
// SIG // VVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcT
// SIG // B1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
// SIG // b3JhdGlvbjEoMCYGA1UEAxMfTWljcm9zb2Z0IENvZGUg
// SIG // U2lnbmluZyBQQ0EgMjAxMTCCAiIwDQYJKoZIhvcNAQEB
// SIG // BQADggIPADCCAgoCggIBAKvw+nIQHC6t2G6qghBNNLry
// SIG // tlghn0IbKmvpWlCquAY4GgRJun/DDB7dN2vGEtgL8DjC
// SIG // mQawyDnVARQxQtOJDXlkh36UYCRsr55JnOloXtLfm1Oy
// SIG // CizDr9mpK656Ca/XllnKYBoF6WZ26DJSJhIv56sIUM+z
// SIG // RLdd2MQuA3WraPPLbfM6XKEW9Ea64DhkrG5kNXimoGMP
// SIG // LdNAk/jj3gcN1Vx5pUkp5w2+oBN3vpQ97/vjK1oQH01W
// SIG // KKJ6cuASOrdJXtjt7UORg9l7snuGG9k+sYxd6IlPhBry
// SIG // oS9Z5JA7La4zWMW3Pv4y07MDPbGyr5I4ftKdgCz1TlaR
// SIG // ITUlwzluZH9TupwPrRkjhMv0ugOGjfdf8NBSv4yUh7zA
// SIG // IXQlXxgotswnKDglmDlKNs98sZKuHCOnqWbsYR9q4ShJ
// SIG // nV+I4iVd0yFLPlLEtVc/JAPw0XpbL9Uj43BdD1FGd7P4
// SIG // AOG8rAKCX9vAFbO9G9RVS+c5oQ/pI0m8GLhEfEXkwcNy
// SIG // euBy5yTfv0aZxe/CHFfbg43sTUkwp6uO3+xbn6/83bBm
// SIG // 4sGXgXvt1u1L50kppxMopqd9Z4DmimJ4X7IvhNdXnFy/
// SIG // dygo8e1twyiPLI9AN0/B4YVEicQJTMXUpUMvdJX3bvh4
// SIG // IFgsE11glZo+TzOE2rCIF96eTvSWsLxGoGyY0uDWiIwL
// SIG // AgMBAAGjggHtMIIB6TAQBgkrBgEEAYI3FQEEAwIBADAd
// SIG // BgNVHQ4EFgQUSG5k5VAF04KqFzc3IrVtqMp1ApUwGQYJ
// SIG // KwYBBAGCNxQCBAweCgBTAHUAYgBDAEEwCwYDVR0PBAQD
// SIG // AgGGMA8GA1UdEwEB/wQFMAMBAf8wHwYDVR0jBBgwFoAU
// SIG // ci06AjGQQ7kUBU7h6qfHMdEjiTQwWgYDVR0fBFMwUTBP
// SIG // oE2gS4ZJaHR0cDovL2NybC5taWNyb3NvZnQuY29tL3Br
// SIG // aS9jcmwvcHJvZHVjdHMvTWljUm9vQ2VyQXV0MjAxMV8y
// SIG // MDExXzAzXzIyLmNybDBeBggrBgEFBQcBAQRSMFAwTgYI
// SIG // KwYBBQUHMAKGQmh0dHA6Ly93d3cubWljcm9zb2Z0LmNv
// SIG // bS9wa2kvY2VydHMvTWljUm9vQ2VyQXV0MjAxMV8yMDEx
// SIG // XzAzXzIyLmNydDCBnwYDVR0gBIGXMIGUMIGRBgkrBgEE
// SIG // AYI3LgMwgYMwPwYIKwYBBQUHAgEWM2h0dHA6Ly93d3cu
// SIG // bWljcm9zb2Z0LmNvbS9wa2lvcHMvZG9jcy9wcmltYXJ5
// SIG // Y3BzLmh0bTBABggrBgEFBQcCAjA0HjIgHQBMAGUAZwBh
// SIG // AGwAXwBwAG8AbABpAGMAeQBfAHMAdABhAHQAZQBtAGUA
// SIG // bgB0AC4gHTANBgkqhkiG9w0BAQsFAAOCAgEAZ/KGpZjg
// SIG // VHkaLtPYdGcimwuWEeFjkplCln3SeQyQwWVfLiw++MNy
// SIG // 0W2D/r4/6ArKO79HqaPzadtjvyI1pZddZYSQfYtGUFXY
// SIG // DJJ80hpLHPM8QotS0LD9a+M+By4pm+Y9G6XUtR13lDni
// SIG // 6WTJRD14eiPzE32mkHSDjfTLJgJGKsKKELukqQUMm+1o
// SIG // +mgulaAqPyprWEljHwlpblqYluSD9MCP80Yr3vw70L01
// SIG // 724lruWvJ+3Q3fMOr5kol5hNDj0L8giJ1h/DMhji8MUt
// SIG // zluetEk5CsYKwsatruWy2dsViFFFWDgycScaf7H0J/je
// SIG // LDogaZiyWYlobm+nt3TDQAUGpgEqKD6CPxNNZgvAs031
// SIG // 4Y9/HG8VfUWnduVAKmWjw11SYobDHWM2l4bf2vP48hah
// SIG // mifhzaWX0O5dY0HjWwechz4GdwbRBrF1HxS+YWG18NzG
// SIG // GwS+30HHDiju3mUv7Jf2oVyW2ADWoUa9WfOXpQlLSBCZ
// SIG // gB/QACnFsZulP0V3HjXG0qKin3p6IvpIlR+r+0cjgPWe
// SIG // +L9rt0uX4ut1eBrs6jeZeRhL/9azI2h15q/6/IvrC4Dq
// SIG // aTuv/DDtBEyO3991bWORPdGdVk5Pv4BXIqF4ETIheu9B
// SIG // CrE/+6jMpF3BoYibV3FWTkhFwELJm3ZbCoBIa/15n8G9
// SIG // bW1qyVJzEw16UM0xghn2MIIZ8gIBATCBlTB+MQswCQYD
// SIG // VQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
// SIG // A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
// SIG // IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQg
// SIG // Q29kZSBTaWduaW5nIFBDQSAyMDExAhMzAAADe6EKPstm
// SIG // 6QHAAAAAAAN7MA0GCWCGSAFlAwQCAQUAoIGuMBkGCSqG
// SIG // SIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQBgjcC
// SIG // AQsxDjAMBgorBgEEAYI3AgEVMC8GCSqGSIb3DQEJBDEi
// SIG // BCCk0LbOeo0Jl3UzqyYcR3jw0T+iGBzKrXSBXSbuitil
// SIG // gzBCBgorBgEEAYI3AgEMMTQwMqAUgBIATQBpAGMAcgBv
// SIG // AHMAbwBmAHShGoAYaHR0cDovL3d3dy5taWNyb3NvZnQu
// SIG // Y29tMA0GCSqGSIb3DQEBAQUABIIBgDtcjoyKuHNu9hIs
// SIG // Tdwg9y73LBD15WHlVfW77zi/XhSfjsFfk6j4RjQZV7pR
// SIG // mOG2coM7WfztP3TZS8QkaazYmZYs9zypU9yOl+4LaxGz
// SIG // kggp0hHawYOBcAu+/WWzTgFxZvmiee4igvBW4wmMFRt4
// SIG // nYtlNupp5ScQlmsMnPv1I9qDCL0W9yprXLh/oevTxcOE
// SIG // AVKyxZFn+Ld19LlezGFpX6pAX8EVns79vbNdUmEpDXxF
// SIG // LaCXjcEwEof81ND25i/g9BRHpchHFih28wYWADkrwfP4
// SIG // cVi4WPU/MS/dzzbFoeBfyEKpzj5ySOasEZ61PBvKM2b6
// SIG // f0sHIfPaCeYIYVv0pmXTmQchWwJvAWnqIYGGOXOtS3t7
// SIG // WsRBc4ZQ7fJUu8rciIDfya899bmPk2lK8a4PNQs6yu1k
// SIG // A23V9PXNC+xkt/ViHe585+O3QFiU0LOkpNDlT3UuO8FL
// SIG // E4162XAhqedlRJhdVHNFGjAfh+tAowqRDvmPJJCCFrZi
// SIG // rwdWNZhpHmtlvKGCFwAwghb8BgorBgEEAYI3AwMBMYIW
// SIG // 7DCCFugGCSqGSIb3DQEHAqCCFtkwghbVAgEDMQ8wDQYJ
// SIG // YIZIAWUDBAIBBQAwggFRBgsqhkiG9w0BCRABBKCCAUAE
// SIG // ggE8MIIBOAIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFl
// SIG // AwQCAQUABCAYLHeg2MtRToqCzlriiNu+L+ABybJR7PFh
// SIG // n5uhY3gtigIGZK/5D29pGBMyMDIzMDcxODAyMjc1Ni40
// SIG // NjFaMASAAgH0oIHQpIHNMIHKMQswCQYDVQQGEwJVUzET
// SIG // MBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
// SIG // bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0
// SIG // aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1lcmljYSBP
// SIG // cGVyYXRpb25zMSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVT
// SIG // TjpENkJELUUzRTctMTY4NTElMCMGA1UEAxMcTWljcm9z
// SIG // b2Z0IFRpbWUtU3RhbXAgU2VydmljZaCCEVcwggcMMIIE
// SIG // 9KADAgECAhMzAAABx/sAoEpb8ifcAAEAAAHHMA0GCSqG
// SIG // SIb3DQEBCwUAMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQI
// SIG // EwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4w
// SIG // HAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAk
// SIG // BgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAy
// SIG // MDEwMB4XDTIyMTEwNDE5MDEzNVoXDTI0MDIwMjE5MDEz
// SIG // NVowgcoxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNo
// SIG // aW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQK
// SIG // ExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJTAjBgNVBAsT
// SIG // HE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJhdGlvbnMxJjAk
// SIG // BgNVBAsTHVRoYWxlcyBUU1MgRVNOOkQ2QkQtRTNFNy0x
// SIG // Njg1MSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFt
// SIG // cCBTZXJ2aWNlMIICIjANBgkqhkiG9w0BAQEFAAOCAg8A
// SIG // MIICCgKCAgEAr0LcVtnatNFMBrQTtG9P8ISAPyyGmxNf
// SIG // hEzaOVlt088pBUFAIasmN/eOijE6Ucaf3c2bVnN/02ih
// SIG // 0smSqYkm5P3ZwU7ZW202b6cPDJjXcrjJj0qfnuccBtE3
// SIG // WU0vZ8CiQD7qrKxeF8YBNcS+PVtvsqhd5YW6AwhWqhjw
// SIG // 1mYuLetF5b6aPif/3RzlyqG3SV7QPiSJends7gG435Rs
// SIG // y1HJ4XnqztOJR41I0j3EQ05JMF5QNRi7kT6vXTT+MHVj
// SIG // 27FVQ7bef/U+2EAbFj2X2AOWbvglYaYnM3m/I/OWDHUg
// SIG // Gw8KIdsDh3W1eusnF2D7oenGgtahs+S1G5Uolf5ESg/9
// SIG // Z+38rhQwLgokY5k6p8k5arYWtszdJK6JiIRl843H74k7
// SIG // +QqlT2LbAQPq8ivQv0gdclW2aJun1KrW+v52R3vAHCOt
// SIG // bUmxvD1eNGHqGqLagtlq9UFXKXuXnqXJqruCYmfwdFMD
// SIG // 0UP6ii1lFdeKL87PdjdAwyCiVcCEoLnvDzyvjNjxtkTd
// SIG // z6R4yF1N/X4PSQH4FlgslyBIXggaSlPtvPuxAtuac/IT
// SIG // j4k0IRShGiYLBM2Dw6oesLOoxe07OUPO+qXXOcJMVHhE
// SIG // 0MlhhnxfN2B1JWFPWwQ6ooWiqAOQDqzcDx+79shxA1Cx
// SIG // 0K70eOBplMog27gYoLpBv7nRz4tHqoTyvA0CAwEAAaOC
// SIG // ATYwggEyMB0GA1UdDgQWBBQFUNLdHD7BAF/VU/X/eEHL
// SIG // iUSSIDAfBgNVHSMEGDAWgBSfpxVdAF5iXYP05dJlpxtT
// SIG // NRnpcjBfBgNVHR8EWDBWMFSgUqBQhk5odHRwOi8vd3d3
// SIG // Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NybC9NaWNyb3Nv
// SIG // ZnQlMjBUaW1lLVN0YW1wJTIwUENBJTIwMjAxMCgxKS5j
// SIG // cmwwbAYIKwYBBQUHAQEEYDBeMFwGCCsGAQUFBzAChlBo
// SIG // dHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2Nl
// SIG // cnRzL01pY3Jvc29mdCUyMFRpbWUtU3RhbXAlMjBQQ0El
// SIG // MjAyMDEwKDEpLmNydDAMBgNVHRMBAf8EAjAAMBMGA1Ud
// SIG // JQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBCwUAA4IC
// SIG // AQDQy5c8ogP0y8xAsLVca07wWy1mT+nqYgAFnz2972kN
// SIG // O+KJ7AE4f+SVbvOnkeeuOPq3xc+6TS8g3FuKKYEwYqvn
// SIG // RHxX58tjlscZsZeKnu7fGNUlpNT9bOQFHWALURuoXp8T
// SIG // LHhxj3PEq9jzFYBP2YNMLol70ojY1qpze3nMMJfpdurd
// SIG // BBpaOLlJmRNTLhxd+RJGJQbY1XAcx6p/FigwqBasSDUx
// SIG // p+0yFPEBB9uBE3KILAtq6fczGp4EMeon6YmkyCGAtXMK
// SIG // DFQQgdP/ITe7VghAVbPTVlP3hY1dFgc+t8YK2obFSFVK
// SIG // slkASATDHulCMht+WrIsukclEUP9DaMmpq7S0RLODMic
// SIG // I6PtqqGOhdnaRltA0d+Wf+0tPt9SUVtrPJyO7WMPKbyk
// SIG // CRXzmHK06zr0kn1YiUYNXCsOgaHF5ImO2ZwQ54UE1I55
// SIG // jjUdldyjy/UPJgxRm9NyXeO7adYr8K8f6Q2nPF0vWqFG
// SIG // 7ewwaAl5ClKerzshfhB8zujVR0d1Ra7Z01lnXYhWuPqV
// SIG // ZayFl7JHr6i6huhpU6BQ6/VgY0cBiksX4mNM+ISY81T1
// SIG // RYt7fWATNu/zkjINczipzbfg5S+3fCAo8gVB6+6A5L0v
// SIG // Bg39dsFITv6MWJuQ8ZZy7fwlFBZE4d5IFbRudakNwKGd
// SIG // yLGM2otaNq7wm3ku7x41UGAmkDCCB3EwggVZoAMCAQIC
// SIG // EzMAAAAVxedrngKbSZkAAAAAABUwDQYJKoZIhvcNAQEL
// SIG // BQAwgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNo
// SIG // aW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQK
// SIG // ExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xMjAwBgNVBAMT
// SIG // KU1pY3Jvc29mdCBSb290IENlcnRpZmljYXRlIEF1dGhv
// SIG // cml0eSAyMDEwMB4XDTIxMDkzMDE4MjIyNVoXDTMwMDkz
// SIG // MDE4MzIyNVowfDELMAkGA1UEBhMCVVMxEzARBgNVBAgT
// SIG // Cldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAc
// SIG // BgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQG
// SIG // A1UEAxMdTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIw
// SIG // MTAwggIiMA0GCSqGSIb3DQEBAQUAA4ICDwAwggIKAoIC
// SIG // AQDk4aZM57RyIQt5osvXJHm9DtWC0/3unAcH0qlsTnXI
// SIG // yjVX9gF/bErg4r25PhdgM/9cT8dm95VTcVrifkpa/rg2
// SIG // Z4VGIwy1jRPPdzLAEBjoYH1qUoNEt6aORmsHFPPFdvWG
// SIG // UNzBRMhxXFExN6AKOG6N7dcP2CZTfDlhAnrEqv1yaa8d
// SIG // q6z2Nr41JmTamDu6GnszrYBbfowQHJ1S/rboYiXcag/P
// SIG // XfT+jlPP1uyFVk3v3byNpOORj7I5LFGc6XBpDco2LXCO
// SIG // Mcg1KL3jtIckw+DJj361VI/c+gVVmG1oO5pGve2krnop
// SIG // N6zL64NF50ZuyjLVwIYwXE8s4mKyzbnijYjklqwBSru+
// SIG // cakXW2dg3viSkR4dPf0gz3N9QZpGdc3EXzTdEonW/aUg
// SIG // fX782Z5F37ZyL9t9X4C626p+Nuw2TPYrbqgSUei/BQOj
// SIG // 0XOmTTd0lBw0gg/wEPK3Rxjtp+iZfD9M269ewvPV2HM9
// SIG // Q07BMzlMjgK8QmguEOqEUUbi0b1qGFphAXPKZ6Je1yh2
// SIG // AuIzGHLXpyDwwvoSCtdjbwzJNmSLW6CmgyFdXzB0kZSU
// SIG // 2LlQ+QuJYfM2BjUYhEfb3BvR/bLUHMVr9lxSUV0S2yW6
// SIG // r1AFemzFER1y7435UsSFF5PAPBXbGjfHCBUYP3irRbb1
// SIG // Hode2o+eFnJpxq57t7c+auIurQIDAQABo4IB3TCCAdkw
// SIG // EgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEEAYI3FQIE
// SIG // FgQUKqdS/mTEmr6CkTxGNSnPEP8vBO4wHQYDVR0OBBYE
// SIG // FJ+nFV0AXmJdg/Tl0mWnG1M1GelyMFwGA1UdIARVMFMw
// SIG // UQYMKwYBBAGCN0yDfQEBMEEwPwYIKwYBBQUHAgEWM2h0
// SIG // dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvRG9j
// SIG // cy9SZXBvc2l0b3J5Lmh0bTATBgNVHSUEDDAKBggrBgEF
// SIG // BQcDCDAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMAQTAL
// SIG // BgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAfBgNV
// SIG // HSMEGDAWgBTV9lbLj+iiXGJo0T2UkFvXzpoYxDBWBgNV
// SIG // HR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1pY3Jvc29m
// SIG // dC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNSb29DZXJB
// SIG // dXRfMjAxMC0wNi0yMy5jcmwwWgYIKwYBBQUHAQEETjBM
// SIG // MEoGCCsGAQUFBzAChj5odHRwOi8vd3d3Lm1pY3Jvc29m
// SIG // dC5jb20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dF8yMDEw
// SIG // LTA2LTIzLmNydDANBgkqhkiG9w0BAQsFAAOCAgEAnVV9
// SIG // /Cqt4SwfZwExJFvhnnJL/Klv6lwUtj5OR2R4sQaTlz0x
// SIG // M7U518JxNj/aZGx80HU5bbsPMeTCj/ts0aGUGCLu6WZn
// SIG // OlNN3Zi6th542DYunKmCVgADsAW+iehp4LoJ7nvfam++
// SIG // Kctu2D9IdQHZGN5tggz1bSNU5HhTdSRXud2f8449xvNo
// SIG // 32X2pFaq95W2KFUn0CS9QKC/GbYSEhFdPSfgQJY4rPf5
// SIG // KYnDvBewVIVCs/wMnosZiefwC2qBwoEZQhlSdYo2wh3D
// SIG // YXMuLGt7bj8sCXgU6ZGyqVvfSaN0DLzskYDSPeZKPmY7
// SIG // T7uG+jIa2Zb0j/aRAfbOxnT99kxybxCrdTDFNLB62FD+
// SIG // CljdQDzHVG2dY3RILLFORy3BFARxv2T5JL5zbcqOCb2z
// SIG // AVdJVGTZc9d/HltEAY5aGZFrDZ+kKNxnGSgkujhLmm77
// SIG // IVRrakURR6nxt67I6IleT53S0Ex2tVdUCbFpAUR+fKFh
// SIG // bHP+CrvsQWY9af3LwUFJfn6Tvsv4O+S3Fb+0zj6lMVGE
// SIG // vL8CwYKiexcdFYmNcP7ntdAoGokLjzbaukz5m/8K6TT4
// SIG // JDVnK+ANuOaMmdbhIurwJ0I9JZTmdHRbatGePu1+oDEz
// SIG // fbzL6Xu/OHBE0ZDxyKs6ijoIYn/ZcGNTTY3ugm2lBRDB
// SIG // cQZqELQdVTNYs6FwZvKhggLOMIICNwIBATCB+KGB0KSB
// SIG // zTCByjELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
// SIG // bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoT
// SIG // FU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjElMCMGA1UECxMc
// SIG // TWljcm9zb2Z0IEFtZXJpY2EgT3BlcmF0aW9uczEmMCQG
// SIG // A1UECxMdVGhhbGVzIFRTUyBFU046RDZCRC1FM0U3LTE2
// SIG // ODUxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1w
// SIG // IFNlcnZpY2WiIwoBATAHBgUrDgMCGgMVAOIASP0JSbv5
// SIG // R23wxciQivHyckYooIGDMIGApH4wfDELMAkGA1UEBhMC
// SIG // VVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcT
// SIG // B1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
// SIG // b3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0IFRpbWUt
// SIG // U3RhbXAgUENBIDIwMTAwDQYJKoZIhvcNAQEFBQACBQDo
// SIG // YGWzMCIYDzIwMjMwNzE4MDkxMzIzWhgPMjAyMzA3MTkw
// SIG // OTEzMjNaMHcwPQYKKwYBBAGEWQoEATEvMC0wCgIFAOhg
// SIG // ZbMCAQAwCgIBAAICIGQCAf8wBwIBAAICEk4wCgIFAOhh
// SIG // tzMCAQAwNgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGE
// SIG // WQoDAqAKMAgCAQACAwehIKEKMAgCAQACAwGGoDANBgkq
// SIG // hkiG9w0BAQUFAAOBgQABOBP+iyjiMrlRKZFYCg5Tv9ng
// SIG // QSFVWuO4+w6ftTb/SG333l0PDC18+aYUBSQTSpx0+HfS
// SIG // xC174jqhj1rbSqmT+za9z1CGre2MUb5zSBNP/ZX87uyb
// SIG // DaMhZPuy968GHoGoUS+CG4L3mh0Ua6XhU6MOjgHAqHUu
// SIG // r8bc+wDBRUj99zGCBA0wggQJAgEBMIGTMHwxCzAJBgNV
// SIG // BAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
// SIG // VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
// SIG // Q29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBU
// SIG // aW1lLVN0YW1wIFBDQSAyMDEwAhMzAAABx/sAoEpb8ifc
// SIG // AAEAAAHHMA0GCWCGSAFlAwQCAQUAoIIBSjAaBgkqhkiG
// SIG // 9w0BCQMxDQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkE
// SIG // MSIEIOPwqQsaAl7lcWFLQReh7egCdyTRqa4r4/Tmmrv/
// SIG // YXUlMIH6BgsqhkiG9w0BCRACLzGB6jCB5zCB5DCBvQQg
// SIG // R+fl2+JSskULOeVYLbeMgk7HdIbREmAsjwtcy6MJkskw
// SIG // gZgwgYCkfjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
// SIG // V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwG
// SIG // A1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYD
// SIG // VQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAx
// SIG // MAITMwAAAcf7AKBKW/In3AABAAABxzAiBCB2aRqRlw/w
// SIG // XzewLppjRX0x9aWCCGCkI8sq51xNZFStpDANBgkqhkiG
// SIG // 9w0BAQsFAASCAgCPaXVkkzFRXDB5z+Es5DqWmaqXaI0P
// SIG // oZ4nFnAQ6gwVFTvi80q6pkgzLi8+D1AJ4mLGawynoQ7z
// SIG // dxVmSKQWf1VHcMVGoPsHuNlEJGoEuvFY4GOlt5ZXE/B0
// SIG // 4xBPgaYTwxtWDCC/nFP6bhdvweIjjQf1oHpTxRnwbvmb
// SIG // xRBprNm8XdbTeKl9HDIJuPevkH/d+/KmDhYStWokpYa5
// SIG // TCAM8AwUfo4mc/68Lz20Op9O1/cqMM4oJUh9C2tVVhNS
// SIG // EB9mtUdrDNKsuJr/4VBjFoCAlZeSjzGlgxIf4PH282MV
// SIG // wxwx3dD9ehRY5FR854tPqEpqoQoj2ghCSX+lTZQ/ZbTc
// SIG // S57ui9FSa576ZDeJawnGepXy8Nx6drNIktk1Hxw56h5S
// SIG // XnSZAqAtjOdDueigaj3uFT60+z9udijpeYTXbQsw74/v
// SIG // lOifmZk2IZ2kkKsD5c28jyTWBgDM9MEvrmKgr4Z9ktS1
// SIG // F/m8IFWIN4trinMRiJvT0Dy+IWk+SNcV8AxWU4OovNch
// SIG // hy4bFFnWG6ccOBZT0DUAk39dZjHTgjNZNY5rkSYynlSk
// SIG // O81tX+1kq2Ms1zJPOBHqy6A/AGWiehYezoLoZWPtajzd
// SIG // AyePIrFzHhyk1Eu3evBQMXUOrTTrGJ0oXohlErM8s5b5
// SIG // znUPRsFLX8ECiFjxFmoqVl+RO3MPWwWwwomzWQ==
// SIG // End signature block
